"""XRPC server implementation."""
from functools import wraps
import logging

from .base import fail, METHOD_TYPES, NSID_RE, Base

logger = logging.getLogger(__name__)


class Redirect(Exception):
    """Raised by XRPC handlers to direct the server to serve an HTTP redirect.

    Whether this is official supported by the XRPC spec is still TBD:
    https://github.com/bluesky-social/atproto/discussions/1228

    Attributes:
      to (str): URL to redirect to
      status (int): HTTP status code, defaults to 302
    """
    def __init__(self, to, status=302, headers=None):
        assert to
        assert status
        self.to = to
        self.status = status
        self.headers = headers or {}


class Server(Base):
    """XRPC server base class. Subclass to implement specific methods."""
    _methods = None  # dict, maps string NSID to Python callable

    def __init__(self, **kwargs):
        """Constructor.

        Args:
          kwargs: passed through to :class:`Base`

        Raises:
          ValidationError: if any schema is invalid
        """
        super().__init__(**kwargs)
        self._methods = {}

    def method(self, nsid):
        """XRPC method decorator. Use on each function that implements a method.

        Args:
          nsid (str)
        """
        def decorated(fn):
            self.register(nsid, fn)

            @wraps(fn)
            def wrapped(*args, **kwargs):
                return fn(*args, **kwargs)
            return wrapped

        return decorated

    def register(self, nsid, fn):
        """Registers an XRPC method decorator. Alternative to :meth:`method`.

        Args:
          nsid (str)
          fn (callable)
        """
        assert NSID_RE.match(nsid)

        existing = self._methods.get(nsid)
        if existing:
            fail(f'{nsid} already registered with {existing}',  AssertionError)
        self._methods[nsid] = fn

    def call(self, nsid, input=None, **params):
        """Calls an XRPC query or procedure method.

        For subscriptions, returns a generator that yields ``(header dict, payload
        dict)`` tuples to be DAG-CBOR encoded and sent to the websocket client.

        Args:
          nsid (str): method NSID
          input (dict or bytes): input body, optional for subscriptions
          params: optional parameters

        Returns:
          dict: output

        Raises:
          NotImplementedError: if the given NSID is not implemented or found in
            any of the loaded lexicons
          ValidationError: if the parameters, input, or returned output don't
            validate against the method's schemas
        """
        logger.debug(f'{nsid}: {params} {self.loggable(input)}')

        fn = self._methods.get(nsid)
        if not fn:
            fail(f'{nsid} not implemented', NotImplementedError)

        subscription = self.defs[nsid]['type'] == 'subscription'

        # validate params and input, then encode params
        params = self.validate(nsid, 'parameters', params)
        input = self.validate(nsid, 'input', input)

        args = [] if subscription else [input]
        output = fn(*args, **params)

        if subscription:
            def validator():
                for header, payload in output:
                    payload = self.validate(nsid, 'message', payload)
                    yield header, payload
            return validator()
        else:
            logger.debug(f'Returning {self.loggable(output)}')
            return self.validate(nsid, 'output', output)



"""XRPC client implementation.

TODO:

* asyncio support for subscription websockets
"""
import copy
from io import BytesIO, IOBase
import json
import logging
from urllib.parse import urljoin

import libipld
import requests
import simple_websocket

from .base import Base, NSID_SEGMENT_RE

logger = logging.getLogger(__name__)

DEFAULT_PDS = 'https://bsky.social/'
DEFAULT_HEADERS = {
    'User-Agent': 'lexrpc (https://lexrpc.readthedocs.io/)',
}
LOGIN_NSID = 'com.atproto.server.createSession'
REFRESH_NSID = 'com.atproto.server.refreshSession'
TOKEN_ERRORS = (
    'AccountNotFound',
    'AuthenticationRequired',
    'ExpiredToken',
    'InvalidToken',
    'TokenRequired',
)


class _NsidClient():
    """Internal helper class to implement dynamic attribute-based method calls.

    eg ``client.com.example.my_method(...)``
    """
    client = None
    nsid = None

    def __init__(self, client, nsid):
        assert client and nsid
        self.client = client
        self.nsid = nsid

    def __getattr__(self, attr):
        segment = attr.replace('_', '-')
        if NSID_SEGMENT_RE.match(segment):
            return _NsidClient(self.client, f'{self.nsid}.{segment}')

        return getattr(super(), attr)

    def __call__(self, *args, **kwargs):
        return self.client.call(self.nsid, *args, **kwargs)


class Client(Base):
    """XRPC client.

    Calling ``com.atproto.server.createSession`` will store the returned session
    and include its acccess token in subsequent requests. If a request fails
    with ``ExpiredToken`` and we have a session stored, the access token will be
    refreshed with ``com.atproto.server.refreshSession`` and then the original
    request will be retried.

    Attributes:
      address (str): base URL of XRPC server, eg ``https://bsky.social/``
      session (dict): ``createSession`` response with ``accessJwt``,
        `refreshJwt``, ``handle``, and ``did``
      headers (dict): HTTP headers to include in every request
      auth (requests.auth.AuthBase): optional, used to authenticate requests
    """

    def __init__(self, address=DEFAULT_PDS, access_token=None,
                 refresh_token=None, headers=None, session_callback=None,
                 auth=None, **kwargs):
        """Constructor.

        Args:
          address (str): base URL of XRPC server, eg ``https://bsky.social/``
          access_token (str): optional, will be sent in ``Authorization`` header
          refresh_token (str): optional; used to refresh access token
          headers (dict): optional, HTTP headers to include in every request
          session_callback (callable, dict => None): called when a new session
            is created with new access and refresh tokens, or when ``auth.token``
            changes, eg it gets refreshed. This callable is passed one positional
            argument: if the client has ``access_token``, the dict JSON output
            from ``com.atproto.server.createSession`` or
            ``com.atproto.server.refreshSession``; or if the client has ``auth``,
            ``auth itself.
          kwargs: passed through to :class:`Base`

        Raises:
          ValidationError: if any lexicon schema is invalid
        """
        super().__init__(**kwargs)

        assert not ((access_token or refresh_token) and auth)

        # logger.debug(f'Using server at {address}')
        assert address.startswith('http://') or address.startswith('https://'), \
            f"{address} doesn't start with http:// or https://"
        self.address = address
        self.headers = headers or {}
        self.auth = auth
        self.session = {}
        if access_token or refresh_token:
            self.session.update({
                'accessJwt': access_token,
                'refreshJwt': refresh_token,
            })
        self.session_callback = session_callback

    def __getattr__(self, attr):
        if NSID_SEGMENT_RE.match(attr):
            return _NsidClient(self, attr)

        return getattr(super(), attr)

    def call(self, nsid, input=None, headers={}, decode=True, **params):
        """Makes a remote XRPC method call.

        Args:
          nsid (str): method NSID
          input (dict or bytes): input body, optional for subscriptions
          headers (dict): HTTP headers to include in this request. Overrides any
            headers passed to the constructor.
          decode (bool): if this is a subscription, decode header and payload
            before returning, otherwise return raw bytes
          params: optional method parameters

        Returns:
          dict or generator iterator: for queries and procedures, decoded JSON
          object, or None if the method has no output. For subscriptions,
          generator of messages from server, as (dict header, dict payload) tuple
          if ``decode`` is True, bytes otherwise.

        Raises:
          NotImplementedError: if the given NSID is not found in any of the
            loaded lexicons
          ValidationError: if the parameters, input, or returned output don't
            validate against the method's schemas
          requests.RequestException: if the connection or HTTP request to the
            remote server failed
        """
        # logger.debug(f'{nsid}: {params} {self.loggable(input)}')

        # strip null params, validate params and input, then encode params
        params = {k: v for k, v in params.items() if v is not None}
        params = self.validate(nsid, 'parameters', params)
        params_str = self.encode_params(params)

        type = self._get_def(nsid)['type']
        if type == 'subscription':
            input = self.validate(nsid, 'input', input)

        headers = {
            **DEFAULT_HEADERS,
            'Content-Type': 'application/json',
            **self.headers,
            **headers,
        }

        # auth
        token = (self.session.get('refreshJwt') if nsid == REFRESH_NSID
                 else self.session.get('accessJwt'))
        if token:
            headers['Authorization'] = f'Bearer {token}'

        # run method
        url = urljoin(self.address, f'/xrpc/{nsid}')
        if params_str:
            url += f'?{params_str}'

        # event stream
        if type == 'subscription':
            return self._subscribe(url, nsid, decode=decode)

        # query or procedure
        fn = requests.get if type == 'query' else requests.post

        # buffer binary inputs in memory. ideally we'd stream instead, but if we
        # have to refresh our token below, we need to seek the stream back to the
        # beginning, and not all streams are seekable, eg requests.Request.raw
        if isinstance(input, IOBase) or hasattr(input, 'read'):
            input = input.read()

        logger.debug(f'requests.{fn} {url} {params_str} {self.loggable(input)} {headers} {self.auth}')

        if self.auth:
            orig_token = getattr(self.auth, 'token', None)
        resp = fn(
          url,
          json=input if input and isinstance(input, dict) else None,
          data=input if input and not isinstance(input, dict) else None,
          headers=headers,
          auth=self.auth,
        )

        if (self.auth and self.session_callback
                and getattr(self.auth, 'token', None) != orig_token):
            self.session_callback(self.auth)

        output = None
        content_type = resp.headers.get('Content-Type', '').split(';')[0]
        if content_type == 'application/json' and resp.content:
            output = resp.json()

        if not resp.ok:
            logger.debug(f'Got {resp.status_code}: {resp.text}')

        if nsid in (LOGIN_NSID, REFRESH_NSID):  # auth
            if resp.ok:
                logger.debug(f'Logged in as {output.get("did")}, storing session')
            else:
                logger.debug(f'Login failed, nulling out session')
                output = {}

            self.session = output
            if self.session_callback:
                self.session_callback(output)

        elif not resp.ok:  # token expired, try to refresh it
            if output and output.get('error') in TOKEN_ERRORS:
                self.call(REFRESH_NSID)
                return self.call(nsid, input=input, headers=headers, **params)  # retry

        resp.raise_for_status()

        # logger.debug(json.dumps(output, indent=2))
        output = self.validate(nsid, 'output', output)
        return output

    def _subscribe(self, url, nsid, decode=True):
        """Connects to a subscription websocket, yields the returned messages.

        Args:
          url (str): websocket URL to connect to
          nsid (str): subscription method NSID
          decode (bool): if True, decodes messages before returning

        Returns:
          (dict header, dict payload) or bytes: tuple of dicts if ``decode`` is
            True, otherwise raw bytes

        Raises:
          ValidationError: if ``decode`` is True and an output payload doesn't
            validate gainst the subscription method's lexicon
        """
        ws = simple_websocket.Client(url, headers={
            **DEFAULT_HEADERS,
            **self.headers,
        })

        try:
            while True:
                msg = ws.receive()
                if decode:
                    header, payload = libipld.decode_dag_cbor_multi(msg)
                    payload = self.validate(nsid, 'message', payload)
                    yield (header, payload)
                else:
                    yield msg
        except simple_websocket.ConnectionClosed as cc:
            logger.debug(cc)



"""Flask handler for ``/xrpc/...`` endpoints."""
from collections import defaultdict, namedtuple
from datetime import datetime, timedelta
import logging

import dag_cbor
import dag_json
from flask import after_this_request, make_response, request
from flask.json import jsonify
from flask.views import View
from flask_sock import Sock
from iterators import TimeoutIterator
import libipld
from multiformats import CID
from simple_websocket import ConnectionClosed
from werkzeug.exceptions import TooManyRequests

from . import base
from .base import NSID_RE, ValidationError
from .server import Redirect

logger = logging.getLogger(__name__)

SUBSCRIPTION_ITERATOR_TIMEOUT = timedelta(seconds=10)

RESPONSE_HEADERS = {
    # wide open CORS to allow client-side apps like https://bsky.app/
    'Access-Control-Allow-Headers': '*',
    'Access-Control-Allow-Methods': '*',
    'Access-Control-Allow-Origin': '*',
}

# maps string NSID to Subscriber
subscribers = defaultdict(list)
Subscriber = namedtuple('Subscriber', ('ip', 'user_agent', 'args', 'start'))


def init_flask(xrpc_server, app):
    """Connects a :class:`lexrpc.Server` to serve ``/xrpc/...`` on a Flask app.

    Args:
      xrpc_server (lexrpc.Server)
      app (flask.Flask)
    """
    logger.info(f'Registering {xrpc_server} with {app}')

    sock = Sock(app)
    for nsid, _ in xrpc_server._methods.items():
        if xrpc_server.defs[nsid]['type'] == 'subscription':
            sock.route(f'/xrpc/{nsid}')(subscription(xrpc_server, nsid))

    app.add_url_rule('/xrpc/<nsid>',
                     view_func=XrpcEndpoint.as_view('xrpc-endpoint', xrpc_server),
                     methods=['GET', 'POST', 'OPTIONS'])


class XrpcEndpoint(View):
    """Handles inbound XRPC query and procedure (but not subscription) methods.

    Attributes:
      server (lexrpc.Server)
    """
    server = None

    def __init__(self, server):
        self.server = server

    def dispatch_request(self, nsid):
        if not NSID_RE.match(nsid):
            return {
                'error': 'InvalidRequest',
                'message': f'{nsid} is not a valid NSID',
            }, 400, RESPONSE_HEADERS
        try:
            lexicon = self.server._get_def(nsid)
        except NotImplementedError as e:
            return {
                'error': 'MethodNotImplemented',
                'message': str(e),
            }, 501, RESPONSE_HEADERS

        if lexicon['type'] == 'subscription':
            return {'message': f'Use websocket for {nsid}, not HTTP'}, 405

        if request.method == 'OPTIONS':
            return '', 200, RESPONSE_HEADERS

        # prepare input
        in_encoding = lexicon.get('input', {}).get('encoding')
        if in_encoding in ('application/json', None):
            input = request.json if request.content_length else {}
        else:
            # binary
            if request.content_type != in_encoding:
                logger.warning(f'expecting input encoding {in_encoding}, request has Content-Type {request.content_type} !')
            input = request.get_data()

        # run method
        try:
            params = self.server.decode_params(nsid, request.args.items(multi=True))
            # TODO: for binary input/output, support streaming with eg
            # io.BufferedReader/Writer?
            output = self.server.call(nsid, input=input, **params)
        except Redirect as r:
            return make_response('', r.status, {'Location': r.to, **r.headers})
        except NotImplementedError as e:
            return {
                'error': 'MethodNotImplemented',
                'message': str(e),
            }, 501, RESPONSE_HEADERS
        except (ValidationError, ValueError) as e:
            if isinstance(e, ValueError):
                logging.debug(f'Method raised', exc_info=True)
            return {
                'error': getattr(e, 'name', 'InvalidRequest'),
                'message': getattr(e, 'message', str(e)),
            }, 400, {**RESPONSE_HEADERS, **getattr(e, 'headers', {})}

        # prepare output
        out_encoding = lexicon.get('output', {}).get('encoding')
        if out_encoding in ('application/json', None):
            return jsonify(output or ''), RESPONSE_HEADERS
        else:
            # binary
            if not isinstance(output, (str, bytes)):
                return {'message': f'Expected str or bytes output to match {out_encoding}, got {output.__class__}'}, 500
            return output, RESPONSE_HEADERS


def subscription(xrpc_server, nsid):
    """Generates websocket handlers for inbound XRPC subscription methods.

    Note that this calls the XRPC method on a *different thread*, so that it can
    block on it there while still periodically checking in the request thread
    that the websocket client is still connected.

    Args:
      xrpc_server (lexrpc.Server)
      nsid (str): XRPC method NSID
    """
    def handle(ws):
        """
        Args:
          ws (wsproto.WSConnection)
        """
        params = xrpc_server.decode_params(nsid, request.args.items(multi=True))

        # use TimeoutIterator here so that we can periodically detect if the
        # client has disconnected. if we don't, we'll tie up this thread forever
        # while we block waiting for results from the XRPC server method, and
        # we'll eventually exhaust the WSGI worker thread pool. background:
        # https://github.com/miguelgrinberg/flask-sock/issues/78
        iter = TimeoutIterator(xrpc_server.call(nsid, **params),
                               timeout=SUBSCRIPTION_ITERATOR_TIMEOUT.total_seconds())
        for result in iter:
            if not ws.connected:
                logger.debug(f'Websocket client disconnected from {nsid}')
                iter.interrupt()
                return
            elif result == iter.get_sentinel():
                continue

            header, payload = result
            # TODO: validate header, payload?

            # log
            seq = payload.get('seq')
            did = payload.get('did') or payload.get('repo')
            commit = payload.get('commit')
            if isinstance(commit, CID):
                commit = f'commit {commit.encode("base32")}'
            # can't DAG-JSON encode payload here? maybe? it hits
            # ValueError: Failed to encode DAG-CBOR. Unknown cbor tag `0`
            # https://console.cloud.google.com/errors/detail/CNzlgrvr2bHuvwE;time=PT1H;refresh=true;locations=global?project=bridgy-federated
            # eg dag_json.encode(payload, dialect="atproto")[:500]
            logger.debug(f'Sending {nsid.split(".")[-1]} {seq} {did} {header.get("t")}')

            # emit!
            try:
                ws.send(dag_cbor.encode(header) + dag_cbor.encode(payload))
            except (ConnectionError, ConnectionClosed, OSError) as err:
                logger.debug(f'Websocket client disconnected from {nsid}: {err}')
                iter.interrupt()
                return

    def track_subscriber(ws):
        # support X-Forwarded-For header:
        # https://cloud.google.com/appengine/docs/flexible/reference/request-headers#app_engine-specific_headers
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Forwarded-For
        if x_forwarded_for := request.headers.get('X-Forwarded-For'):
            ip = x_forwarded_for.split(',')[0]
        else:
            ip = request.remote_addr

        for client in subscribers[nsid]:
            if client.ip == ip:
                logger.debug(f'Rejecting connection, already connected for {nsid}: {ip} {request.user_agent}')
                raise TooManyRequests()

        logger.debug(f'New websocket client for {nsid}: {ip} {request.user_agent}')
        subscriber = Subscriber(ip=ip,
                                user_agent=str(request.user_agent),
                                args=request.args.to_dict(),
                                start=base.now().replace(microsecond=0))
        subscribers[nsid].append(subscriber)

        try:
            handle(ws)
        finally:
            # ideally I'd use Flask's after_this_request instead, but it doesn't
            # guarantee that it'll run if the request raises an uncaught
            # exception. teardown_request does, but it runs on *every* request.
            subscribers[nsid].remove(subscriber)

    return track_subscriber



"""Base code shared by both server and client."""
import copy
from datetime import datetime, timezone
from importlib.resources import files
import json
import logging
import re
import string
from urllib.parse import urlencode, urljoin, urlparse

import grapheme
from multiformats import CID

logger = logging.getLogger(__name__)

METHOD_TYPES = frozenset((
    'query',
    'procedure',
    'subscription',
))
LEXICON_TYPES = METHOD_TYPES | frozenset((
    'object',
    'record',
    'ref',
    'token',
))
PARAMETER_TYPES = frozenset((
    'array',
    'boolean',
    'integer',
    'number',
    'string',
))
# https://atproto.com/specs/lexicon#overview-of-types
FIELD_TYPES = {
    'null': type(None),
    'blob': dict,
    'boolean': bool,
    'cid-link': CID,
    'integer': int,
    'string': str,
    'bytes': bytes,
    'array': list,
    'object': dict,
    # these could be tokens
    # 'ref': dict,
    # 'union': dict,
}

# https://atproto.com/specs/data-model#blob-type
BLOB_DEF = {
    'type': 'record',
    'record': {
        'required': ['ref', 'mimeType', 'size'],
        'properties': {
            'ref': {
                'type': 'ref',
            },
            'mimeType': {
                'type': 'string',
                'minLength': 1,
            },
            'size': {
                'type': 'integer',
                'minimum': 1,
            },
        },
    },
}

# duplicated in bridgy-fed/common.py
DOMAIN_PATTERN = r'([a-z0-9][a-z0-9-]{0,62}(?<!-)\.){1,}[a-z][a-z0-9-]*(?<!-)'
DOMAIN_RE = re.compile(f'^{DOMAIN_PATTERN}$')

# https://atproto.com/specs/nsid
NSID_SEGMENT = '[a-zA-Z0-9-]+'
NSID_SEGMENT_RE = re.compile(f'^{NSID_SEGMENT}$')
NSID_PATTERN = r'(?![0-9])((?!-)[a-z0-9-]{1,63}(?<!-)\.){2,}[a-zA-Z][a-zA-Z0-9]{0,62}'
NSID_RE = re.compile(f'^{NSID_PATTERN}$')

# https://atproto.com/specs/lexicon#string-formats
# https://datatracker.ietf.org/doc/html/rfc5646#section-2.1
LANG_RE = re.compile(r'^(i|[a-z]{2,3})(-[A-Za-z0-9-]+)?$')

# https://atproto.com/specs/record-key
RKEY_RE = re.compile(r'^[A-Za-z0-9._:~-]{1,512}$')

# https://atproto.com/specs/record-key#record-key-type-tid
BASE32_CHARS = string.ascii_lowercase + "234567"
TID_RE = re.compile(rf'^[{BASE32_CHARS}]{{13}}$')

CID_RE = re.compile(r'^[A-Za-z0-9+]{8,}$')

# https://www.w3.org/TR/did-core/#did-syntax
DID_PATTERN = r'did:[a-z]+:[A-Za-z0-9._%:-]{1,2048}(?<!:)'
DID_RE = re.compile(f'^{DID_PATTERN}$')

# https://atproto.com/specs/at-uri-scheme
# NOTE: duplicated in granary.bluesky!
# also see arroba.util.parse_at_uri
_CHARS = 'a-zA-Z0-9-.'
AT_URI_RE = re.compile(rf"""
    ^at://
     (?P<repo>{DID_PATTERN}|{DOMAIN_PATTERN})
      (?:/(?P<collection>{NSID_PATTERN})
       (?:/(?P<rkey>[{_CHARS}:~_]+))?)?
    $""", re.VERBOSE)

# wrapper for datetime.now, lets us mock it out in tests
now = lambda tz=timezone.utc, **kwargs: datetime.now(tz=tz, **kwargs)


class XrpcError(ValueError):
    """A named error in an XRPC call.

    ``name`` is the error, eg ``RepoNotFound`` in ``com.atproto.sync.getRepo``.
    ``message`` is the human-readable string error message.
    """
    def __init__(self, message, name=None, **kwargs):
        super().__init__(message, **kwargs)
        self.name = name
        self.message = message


def load_lexicons(traversable):
    if traversable.is_file():
        lexicons = [json.loads(traversable.read_text())]
    elif traversable.is_dir():
        lexicons = sum((load_lexicons(item) for item in traversable.iterdir()),
                       start=[])

    return lexicons

_bundled_lexicons = load_lexicons(files('lexrpc').joinpath('lexicons'))
logger.info(f'{len(_bundled_lexicons)} lexicons loaded')


def fail(msg, exc=NotImplementedError):
    """Logs an error and raises an exception with the given message."""
    logger.error(msg)
    raise exc(msg)


class ValidationError(ValueError):
    """Raised when an object or XRPC input or output doesn't match its schema."""
    pass


class Base():
    """Base class for both XRPC client and server."""

    defs = None  # dict mapping id to lexicon def
    _validate = None
    _truncate = None

    def __init__(self, lexicons=None, validate=True, truncate=False):
        """Constructor.

        Args:
          lexicons (sequence of dict): lexicons, optional. If not provided,
            defaults to the official, built in ``com.atproto`` and ``app.bsky``
            lexicons.
          validate (bool): whether to validate schemas, parameters, and input
            and output bodies
          truncate (bool): whether to truncate string values that are longer
            than their ``maxGraphemes`` or ``maxLength`` in their lexicon

        Raises:
          ValidationError: if any schema is invalid
        """
        self._validate = validate
        self._truncate = truncate
        self.defs = {}

        if lexicons is None:
            lexicons = _bundled_lexicons

        for i, lexicon in enumerate(copy.deepcopy(lexicons)):
            nsid = lexicon.get('id')
            if not nsid or not isinstance(nsid, str):
                raise ValidationError(f'Lexicon {i} missing or invalid id field')
            elif lexicon.get('lexicon') != 1:
                raise ValidationError(f'{nsid} lexicon field should be 1')
            # logger.debug(f'Loading lexicon {nsid}')

            for name, defn in lexicon.get('defs', {}).items():
                id = nsid if name == 'main' else f'{nsid}#{name}'
                self.defs[id] = defn

                type = defn['type']
                if type not in LEXICON_TYPES | PARAMETER_TYPES:
                    raise ValidationError(f'Bad type for lexicon {id}: {type}')

                if validate:
                    for field in ('input', 'output', 'message', 'parameters',
                                  'record'):
                        if schema := defn.get('field'):
                            if not isinstance(schema, dict):
                                raise ValidationError(f'{nsid} {field} is invalid')
                            elif not isinstance(schema.get('properties'), dict):
                                raise ValidationError(f'{nsid} {field} properties is invalid')

                self.defs[id] = defn

        self.defs['blob'] = BLOB_DEF

        if not self.defs:
            logger.error('No lexicons loaded!')

    def _get_def(self, id):
        """Returns the given lexicon def.

        Raises:
          NotImplementedError: if no def exists for the given id
        """
        # TODO: bring back once the Bluesky appview validates this too
        # https://github.com/bluesky-social/atproto/discussions/1968#discussioncomment-11195092
        # if id.endswith('#main'):
        #     # https://atproto.com/specs/lexicon#:~:text=main%20suffix
        #     raise ValidationError(f'#main suffix not allowed on $type: {id}')

        lexicon = self.defs.get(id)
        if not lexicon:
            fail(f'{id} not found')

        return lexicon

    def validate(self, nsid, type, obj):
        """If configured to do so, validates a ATProto value against its lexicon.

        Returns ``None`` if the object validates, otherwise raises an exception.

        Does nothing if this object was initialized with ``validate=False``.

        Args:
          nsid (str): method NSID
          type (str): ``input``, ``output``, ``parameters``, or ``record``
          obj (dict): JSON object

        Returns:
          dict: obj, either unchanged, or possible a modified copy if
            ``truncate`` is enabled and a string value was too long

        Raises:
          NotImplementedError: if no lexicon exists for the given NSID, or the
            lexicon does not define a schema for the given type
          ValidationError: if the object is invalid
        """
        if not self._validate and not self._truncate:
            return obj

        assert type in ('input', 'output', 'message', 'parameters', 'record'), type

        base = self._get_def(nsid).get(type, {})
        encoding = base.get('encoding')
        if encoding and encoding != 'application/json':
            # binary or other non-JSON data, pass through
            return obj

        schema = base
        if type in ('input', 'output', 'message'):
            schema = base.get('schema')

        if not schema:
            return obj
            # ...or should we fail if obj is non-null? maybe not, since then
            # we'd fail if a query with no params gets requested with any query
            # params at all, eg utm_* tracking params

        self._validate_schema(name=type, val=obj, type_=nsid, lexicon=nsid,
                              schema=schema)

        return obj

    def _validate_schema(self, *, name, val, type_, lexicon, schema):
        """Validates an ATProto value against a lexicon schema.

        Returns ``None`` if the value validates, otherwise raises an exception.

        https://atproto.com/specs/lexicon

        Args:
          name (str): field name
          val: value
          type_ (str): name of type, eg ``integer`` or ``app.bsky.feed.post#replyRef``
          lexicon (str): fully qualified lexicon name that contains this schema,
            eg ``app.bsky.feed.post`` or ``app.bsky.feed.post#replyRef``
          schema (dict): schema to validate against if this is a compound
            object and not a primitive

        Raises:
          ValidationError: if the value is invalid
        """
        # logger.debug(f'@ {name} {type_} {lexicon} {str(val)[:100]} {str(schema)[:100]}')

        def get_schema(lex_name):
            """Returns (fully qualified lexicon name, lexicon) tuple."""
            schema_name = urljoin(lexicon, lex_name)
            schema = self._get_def(schema_name)
            if schema.get('type') == 'record':
                schema = schema.get('record')
            if not schema:
                fail(f'lexicon {schema_name} not found')
            return schema_name, schema

        def fail(msg):
            if self._validate:
                val_str = repr(val)
                if len(val_str) > 50:
                    val_str = val_str[:50] + '…'
                prefix = f'in {lexicon}, ' if lexicon != type_ else ''
                raise ValidationError(
                    f'{prefix}{type_} {name} with value `{val_str}`: {msg}')

        if const := schema.get('const'):
            if val != const:
                fail(f'is not const value {const}')

        if enums := schema.get('enum'):
            if val not in enums:
                fail('is not one of enum values')

        if type_ == 'unknown':
            if isinstance(val, dict) and val.get('$type'):
                lexicon, schema = get_schema(val['$type'])
                # pass through and validate with this schema
            else:
                return

        if expected := FIELD_TYPES.get(type_):
            if type(val) != expected:
                fail(f'has unexpected type {type(val).__name__}')

        if type_ in ('array', 'bytes', 'string'):
            min_length = schema.get('minLength')
            max_length = schema.get('maxLength')
            length = len(val.encode('utf-8') if type_ == 'string' else val)
            if max_length and length > max_length:
                fail(f'is longer ({length}) than maxLength {max_length}')
            elif min_length and length < min_length:
                fail(f'is shorter ({length}) than minLength {min_length}')

        if type_ == 'string':
            if format := schema.get('format'):
                try:
                    self._validate_string_format(val, format)
                except ValidationError as e:
                    fail(e.args[0])

            min_graphemes = schema.get('minGraphemes')
            max_graphemes = schema.get('maxGraphemes')
            if min_graphemes or max_graphemes:
                length = grapheme.length(val)
                if min_graphemes and length < min_graphemes:
                    fail(f'is shorter than minGraphemes {min_graphemes}')
                if max_graphemes and length > max_graphemes:
                    fail(f'is longer than maxGraphemes {max_graphemes}')

        if minimum := schema.get('minimum'):
            if val < minimum:
                fail(f'is lower than minimum {minimum}')
        if maximum := schema.get('maximum'):
            if val > maximum:
                fail(f'is higher than maximum {maximum}')

        if schema and schema.get('type') == 'token':
            if val != lexicon:
                fail(f'is not token {lexicon}')
            elif val not in self.defs:
                fail(f'not found')

        if type_ == 'ref':
            ref = schema['ref']
            if isinstance(val, str) and val != ref:
                fail(f'is not {ref}')
            elif not isinstance(val, dict):
                fail('is not object')
            lexicon, schema = get_schema(ref)

        if type_ == 'union':
            if isinstance(val, dict):
                inner_type = val.get('$type')
                if not inner_type:
                    fail('missing $type')
            elif isinstance(val, str):
                inner_type = val
            else:
                fail("is invalid")

            if schema.get('closed'):
                refs = [urljoin(lexicon, ref) for ref in schema['refs']]
                if inner_type not in refs:
                    fail(f"{inner_type} isn't one of {refs}")

            try:
                lexicon, schema = get_schema(inner_type)
            except NotImplementedError:
                # https://github.com/bluesky-social/atproto/discussions/2940
                # https://github.com/snarfed/lexrpc/issues/16
                logger.debug(f'Skipping unknown type {inner_type}')
                return

        # TODO: maybe bring back once we figure out why the AppView isn't
        # currently enforcing these:
        # https://github.com/snarfed/bridgy-fed/issues/1348#issuecomment-2381056468
        # if type_ == 'blob':
        #     if max_size := schema.get('maxSize'):
        #         # old-style blobs don't have size
        #         # https://atproto.com/specs/data-model#blob-type
        #         if size := val.get('size'):
        #             if size > max_size:
        #                 fail(f'has size {val["size"]} over maxSize {max_size}')
        #     self.validate_mime_type(val['mimeType'], schema.get('accept'), name=name)

        if type_ == 'array':
            for item in val:
                self._validate_schema(
                    name=name, val=item, type_=schema['items']['type'],
                    lexicon=lexicon, schema=schema['items'])

        props = schema.get('properties', {})
        if props and not isinstance(val, dict):
            fail('should be object')

        required = schema.get('required', [])
        nullable = schema.get('nullable', [])
        for prop_name, prop_schema in props.items():
            if prop_name not in val:
                if prop_name in required:
                    fail(f'missing required property {prop_name}')
                continue

            prop_type = prop_schema['type']
            prop_lexicon = lexicon
            prop_val = val[prop_name]
            if prop_val is None:
                if prop_type != 'null' and prop_name not in nullable:
                    fail(f'property {prop_name} is not nullable')
                continue

            elif self._truncate and (max_graphemes := prop_schema.get('maxGraphemes')):
                if grapheme.length(prop_val) > max_graphemes:
                    prop_val = val[prop_name] = grapheme.slice(
                        prop_val, end=max_graphemes - 1) + '…'

            elif prop_type == 'ref':
                prop_lexicon, prop_schema = get_schema(prop_schema['ref'])
                prop_type = prop_schema['type']

            self._validate_schema(name=prop_name, val=prop_val, type_=prop_type,
                                  lexicon=prop_lexicon, schema=prop_schema)

        # unknown parameters aren't allowed
        if schema.get('type') == 'params':
            if unknown := val.keys() - props.keys():
                fail(f'unknown parameters: {unknown}')

    def _validate_string_format(self, val, format):
        """Validates an ATProto string value against a format.

        https://atproto.com/specs/lexicon#string-formats

        Args:
          val (str)
          format (str): one of the ATProto string formats

        Raises:
          ValidationError: if the value is invalid for the given format
        """
        def check(condition):
            if not condition:
                raise ValidationError(f'{val} is invalid for format {format}')

        check(val)

        # TODO: switch to match once we require Python 3.10+
        if format == 'at-identifier':
            check(DID_RE.match(val) or DOMAIN_RE.match(val.lower()))

        elif format == 'at-uri':
            check(len(val) < 8 * 1024)
            check(AT_URI_RE.match(val))
            check('/./' not in val
                  and '/../' not in val
                  and not val.endswith('/.')
                  and not val.endswith('/..'))

        elif format == 'cid':
            # ideally I'd use CID.decode here, but it doesn't support CIDv5,
            # it's too strict about padding, etc.
            check(CID_RE.match(val))

        elif format == 'datetime':
            check('T' in val)

            orig_val = val
            # timezone is required
            val = re.sub(r'([+-][0-9]{2}:[0-9]{2}|Z)$', '', orig_val)
            check(val != orig_val)

            # strip fractional seconds
            val = re.sub(r'\.[0-9]+$', '', val)

            try:
                datetime.fromisoformat(val)
            except ValueError:
                check(False)

        elif format == 'did':
            check(DID_RE.match(val))

        elif format == 'nsid':
            check(len(val) <= 317)
            check(NSID_RE.match(val) and '.' in val)

        elif format in 'handle':
            check(len(val) <= 253)
            check(DOMAIN_RE.match(val.lower()))

        elif format == 'tid':
            check(TID_RE.match(val))
            # high bit, big-endian, can't be 1
            check(not ord(val[0]) & 0x40)

        elif format == 'record-key':
            check(val not in ('.', '..') and RKEY_RE.match(val))

        elif format == 'uri':
            check(len(val) < 8 * 1024)
            check(' ' not in val)
            parsed = urlparse(val)
            check(parsed.scheme
                  and parsed.scheme[0].lower() in string.ascii_lowercase
                  and (parsed.netloc or parsed.path or parsed.query
                       or parsed.fragment))

        elif format == 'language':
            check(LANG_RE.match(val))

        else:
            raise ValidationError(f'unknown format {format}')

    @staticmethod
    def validate_mime_type(mime_type, accept, name=''):
        """Validates that a MIME type matches an accept range.

        For validating blob types. Returns ``None`` if the ``accept`` is empty
        or ``mime_type`` matches, otherwise raises an exception.

        https://atproto.com/specs/lexicon#field-type-definitions

        Args:
          mime_type (str)
          accept (sequence of str): blob ``accept`` field value, list of MIME
            type patterns, eg ``image/jpeg``, ``image/*``, or ``*/*``.
          name: blob field name, only used in exception message

        Raises:
          ValidationError: if ``mime_type`` doesn't match any pattern in ``accept``
        """
        if not accept or '*/*' in accept:
            return

        if not mime_type or (mime_type not in accept
                             and (mime_type.split('/')[0] + '/*') not in accept):
            raise ValidationError(f'blob {name} MIME type {mime_type} not in accept types {accept}')

    def encode_params(self, params):
        """Encodes decoded parameter values.

        Based on https://atproto.com/specs/xrpc#lexicon-http-endpoints

        Args:
          params (dict): maps str names to boolean, number, str, or list values

        Returns:
          bytes: URL-encoded query parameter string
        """
        return urlencode({
            name: ('true' if val is True
                   else 'false' if val is False
                   else val)
            for name, val in params.items()
        }, doseq=True)

    def decode_params(self, method_nsid, params):
        """Decodes encoded parameter values.

        Based on https://atproto.com/specs/xrpc#lexicon-http-endpoints

        Args:
          method_nsid (str):
          params (sequence of (str, str) tuple): name/value mappings

        Returns:
          dict: maps str names to decoded boolean, number, str, and array values

        Raises:
          ValueError: if a parameter value can't be decoded
          NotImplementedError: if no method lexicon is registered for the given NSID
        """
        lexicon = self._get_def(method_nsid)
        params_schema = lexicon.get('parameters', {}).get('properties', {})

        decoded = {}
        for name, val in params:
            type = params_schema.get(name, {}).get('type') or 'string'
            assert type in PARAMETER_TYPES, type

            if type == 'boolean':
                if val == 'true':
                    decoded[name] = True
                elif val == 'false':
                    decoded[name] = False
                else:
                    raise ValueError(
                        f'Got {val!r} for boolean parameter {name}, expected true or false')

            try:
                if type == 'number':
                    decoded[name] = float(val)
                elif type in ('int', 'integer'):
                    decoded[name] = int(val)
            except ValueError as e:
                e.args = [f'{e.args[0]} for {type} parameter {name}']
                raise e

            if type == 'string':
                decoded[name] = val

            if type == 'array':
                decoded.setdefault(name, []).append(val)

        return decoded

    @classmethod
    def loggable(cls, val):
        return (
            f'{len(val)} bytes' if isinstance(val, bytes)
            else val[:100] if isinstance(val, str)
            else [cls.loggable(v) for v in val] if isinstance(val, list)
            else {k: cls.loggable(v) for k, v in val.items()} if isinstance(val, dict)
            else val)
